
""""Trans.py   translates a proposition written with AND,OR,NOT so that
    all occurrences of AND are replaced by OR,NOT.

    Input: a string whose grammatical structure is defined as follows:
            EXPR ::=  VAR  |  ( EXPR OP EXPR )  | ~ EXPR
                 OP is  "^"  or  "v"
                 VAR is an alphabetic letter

    Output: an operator tree built according to this structure:
            optree ::=  VAR  |  ["v", optree, optree ]  | ["~", optree]

            whose truth-table meaning is identical to the meaning
            of the Input.

    Algorithm:
      1. read the Input string and scan it, that is, split it into
         a list of its words.  Call this the  ``wordlist''.
         (Example:   "(P ^ ~Q)" is scanned into
                     wordlist = ["(", "P", "^",  "~", "Q", ")"]  
         )
      2. One by one, move the strings in the wordlist into variable
         nextword  and  call  parseEXPR,  which organizes the strings
         into its operator tree according to the grammar rule for EXPR.
         (Example:  the previous  wordlist  is parsed into this operator
          tree:
                   optree =  ["^", "P", ["~", "Q"]]
         )
     3. Next, replace each occurrence of  ["^", t1, t2]  in optree
        by  ["~", ["v", ["~", t1], ["~", t2]]]
        (Example:
                  optree2 = ["~", ["v", ["~", "P"], ["~", ["~", "Q"]]]]
        )
     4. Finally, erase all double negations:  ["~", ["~", t]]
        is replaced by  t
        (Example:
                  optree3 = ["~", ["v", ["~", "P"], "Q"]]

     5.  The final result prints as  "~(~P v Q)"
"""

"""
SAMPLE EXECUTION:

$ python Trans.py
Type proposition to translate: ( p ^ ( q ^ ~r ))
parsed input: ['^', 'p', ['^', 'q', ['~', 'r']]]
which denotes (p ^ (q ^ ~r))

ands removed:  ['~', ['v', ['~', 'p'], ['~', ['~', ['v', ['~', 'q'], ['~', ['~', 'r']]]]]]]
which denotes ~(~p v ~~(~q v ~~r))

double negations removed: ['~', ['v', ['~', 'p'], ['v', ['~', 'q'], 'r']]]
which denotes ~(~p v (~q v r))
"""

# global variables shared by functions:

nextword = ""  # holds the next unread word from the Input
# global invariant:  nextword + wordlist == all remaining unread words

EOF = "!"      # a word that marks the end of the input words


def scan(text) :
    """scan splits apart the symbols in  text  into a list.

       pre:  text is a string holding a proposition
       post: answer is a list of the words and symbols within  text
             (spaces and newlines are removed)
       returns: answer
    """
    OPERATORS = ("(", ")", "^", "v", "~")  # must be one-symbol operators
    SPACES = (" ", "\n")
    SEPARATORS = OPERATORS + SPACES
 
    nextword = ""  # the next symbol/word to be added to the answer list
    answer = []

    for letter in text:
        # invariant:  answer + nextword + letter  equals all the words/symbols
        #     read so far from  text  with  SPACES  removed

        # see if  nextword  is complete and should be appended to answer:
        if letter in SEPARATORS  and  nextword != "" :
            answer.append(nextword)
            nextword = ""

        if letter in OPERATORS :
            answer.append(letter)
        elif letter in SPACES :
            pass  # discard space
        else :    # build a word or numeral
            nextword = nextword + letter

    if nextword != "" :   
        answer.append(nextword)

    return answer


def getNextword() :
    """moves the front word in  wordlist  to  nextword.
       If wordlist is empty, sets  nextword = EOF
    """
    global nextword, wordlist
    if wordlist == [] :
        nextword = EOF
    else :
        nextword = wordlist[0]
        wordlist = wordlist[1:]


def parseEXPR() :
    """builds an EXPR operator tree from the words in  nextword + wordlist,
          where  EXPR ::=  VAR  |  ( EXPR OP EXPR )  | ~ EXPR
                 OP is  "^"  or  "v"
                 VAR is an alphabetic letter

       pre: global var  nextword  holds the next unread word
       post:  ans  is an operator tree holding the expression that
              was read starting from  nextword.
       returns: ans
    """
    if  nextword.isalpha() :   # a VAR
        ans = nextword        
        getNextword()
    elif nextword == "(" :     # ( EXPR op EXPR ) ?
        getNextword()
        tree1 = parseEXPR()
        op = nextword
        if op == "^"  or  op == "v" :
            getNextword()
            tree2 = parseEXPR()
            if nextword == ")" :
                ans = [op, tree1, tree2]
                getNextword()
            else :
                error("missing )")
        else :
            error("missing operator")
    elif nextword == "~" :
        getNextword()
        tree = parseEXPR()
        ans = ["~", tree]
    else :
        error("illegal symbol to start an expression")

    return ans


def replaceAnd(tree) :
    """replaces all occurrences of  ["^", t1, t2 ]  within  tree
       by  ["~', ["v", ["~", t1], ["~", t2]] ]

      pre: tree is an operator tree
      post:  ans  is a new tree that looks like  tree  with the replacements
      returns: ans
    """
    if isinstance(tree, str) :  # a VAR ?
        ans = tree
    else :  # list of form,  [op, ...]  where  op in ("~","^","v")
        if tree[0] == "~" :
            subans = replaceAnd(tree[1])
            ans = ["~", subans]
        elif tree[0] == "v" :
            ans = ["v", replaceAnd(tree[1]), replaceAnd(tree[2])]
        elif tree[0] == "^" :
            subans1 = replaceAnd(tree[1])
            subans2 = replaceAnd(tree[2])
            ans = ["~", ["v", ["~", subans1], ["~", subans2]]]
        else :
            error("bad optree structure")
    return ans

def removeNeg(tree):
    """replaces all occurrences of  ["~", ["~", t]]  by  t  within tree

       pre:  tree is an operator tree
       post: ans is an operator tree that looks like tree but with
             all double negations removed
       returns: ans
    """
    if isinstance(tree, str) :  # a VAR ?
        ans = tree
    else :  # list of form,  [op, ...]  where  op in ("~","^","v")
        if tree[0] == "~" :
            subtree = tree[1]
            if isinstance(subtree, list) and  subtree[0] == "~" :
                ans = removeNeg(subtree[1])
            else :
                ans = ["~", removeNeg(subtree)]
        elif tree[0] == "v" :
            ans = ["v", removeNeg(tree[1]), removeNeg(tree[2])]
        else :
            error("bad optree structure")
    return ans


def toString(tree):
    """reformats  tree  as a linear string
       pre: tree is an optree
       post: ans is a string that holds the words within tree
       returns: ans
    """
    if isinstance(tree, str) :  # a VAR ?
        ans = tree
    else :  # list of form,  [op, ...]  where  op in ("~","^","v")
        if tree[0] == "~" :
            ans = "~" + toString(tree[1])
        elif tree[0] in ("v", "^") :
            ans = "(" + toString(tree[1]) + " " + tree[0] + " " + toString(tree[2]) + ")"
        else :
            error("bad optree structure")
            ans = "?"
    return ans


def error(message) :
    print message
    

#### MAIN starts here: ####

inputtext = raw_input("Type proposition to translate: ")
# say that  inputtext  holds the text that we must parse into a tree:
wordlist = scan(inputtext)  # holds the remaining unread words

print "wordlist =", wordlist
print

nextword = ""  # holds the first unread word
# global invariant:  nextword + wordlist == all remaining unread words
EOF = "!"      # a word that marks the end of the input words
getNextword()   # call this function to move the first word into  nextword:

optree = parseEXPR()
print "parsed input:", optree
print "which denotes", toString(optree)
print

optree2 = replaceAnd(optree)
print "ands removed: ", optree2
print "which denotes", toString(optree2)
print

optree3 = removeNeg(optree2)
print "double negations removed:", optree3
print "which denotes", toString(optree3)

print
raw_input("Press Enter")
